package gov.va.med.esr.common.batchprocess;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;

import gov.va.med.esr.service.SystemParameterService;
import gov.va.med.fw.batchprocess.AbstractDataQueryIncrementalProcess;
import gov.va.med.fw.batchprocess.DataProcessExecutionContext;
import gov.va.med.fw.batchprocess.DataQueryProcessExecutionContext;
import gov.va.med.fw.persistent.QueryInfo;
import gov.va.med.fw.util.StringUtils;
import gov.va.med.fw.util.ThreadPool;

/**
 * Batch process to transmit Veteran 1095B ACA coverage data to the IRS for previous calendar year
 *
 * WI# 218485
 * Create 7 Sep 2015
 * @author vhaisdfaulkj
 *
 * @version 1.0
 */
public class GenerateIRSTransmissionProcess extends AbstractDataQueryIncrementalProcess {

	private static final int DEFAULT_THREAD_POOL_SIZE = 15;
	private static final int DEFAULT_FETCH_SIZE = 110;
	private static final int DEFAULT_THROTTLE_TASKCOUNT_THREASHOLD = 300;
	private static final int DEFAULT_SPAWN_RETRY_PERIOD = 3000;
	private static final int DEFAULT_JOB_RESULT_UPDATE_INTERVAL = 100;
	public static final String CONTEXT_THREAD_CREATOR = "threadCreator";
	public static final String CONTEXT_TASK_COUNT = "taskCount";
	public static final String CONTEXT_THREAD_POOL = "threadPool";
	public static final String TRANSMISSION_SIZE="IRS Transmission Size";
	private int throttleTaskCountThreshold = DEFAULT_THROTTLE_TASKCOUNT_THREASHOLD;
	private int spawnRetryPeriod = DEFAULT_SPAWN_RETRY_PERIOD;
	private SystemParameterService systemParameterService = null;
	int maxBatchSize;
	private int threadPoolSize = DEFAULT_THREAD_POOL_SIZE;
	private int fetchSize = DEFAULT_FETCH_SIZE;
	private String spawnedTaskId;
	private String calendarYear = null;
	private String startDay = "-01-01";
	private ArrayList<BigDecimal> batch  = null;

	protected void executeProcess(DataProcessExecutionContext context) throws Exception {

		//supported custom args are thread pool size and calendar year, comma delim
		//calendar year to default to prevoius year if not identified, this should be typical
		//override calendar year only if need to start current year transmissions early or bulk retransmit an older year
		this.setThreadPoolSize(DEFAULT_THREAD_POOL_SIZE);
		String args = (String)context.getExecutionArguments();
		if (!StringUtils.isEmpty(args)) {
			String[] params = args.split(",");
			if (params != null && params.length == 2) {
				setThreadPoolSize(Integer.valueOf(params[0].trim()));
				calendarYear = params[1].trim();
			} else {
				setThreadPoolSize(Integer.valueOf(args));
			}
		}

		if (calendarYear == null) {
			Calendar prevYear = Calendar.getInstance();
		    prevYear.add(Calendar.YEAR, -1);
		    calendarYear = Integer.toString(prevYear.get(Calendar.YEAR));
		}
		String startDate = calendarYear + startDay;

		String size = this.getSystemParameterService().getByName(TRANSMISSION_SIZE).getValue();
		maxBatchSize = Integer.valueOf(size);
		setFetchSize(maxBatchSize);
		batch = new ArrayList<BigDecimal>();

		logger.debug("threads: " + getThreadPoolSize());
		logger.debug("year: " + calendarYear);

		this.setParamNames(new String[] {"startDate", "taxYear"});
		this.setParamValues(new String[] {startDate, calendarYear});


		super.executeProcess(context);
	}

	protected List doAcquireData(DataQueryProcessExecutionContext context) throws Exception {
		List acquiredData = null;
		if(this.isIdempotent()) {
			acquiredData = executeQuery(context);
		} else {
			QueryInfo query = context.getCurrentDataQuery().getQuery();
			query.setFetchSize(maxBatchSize);
			getDao().scroll(this, query, context);
		}
		return acquiredData;
	}

	protected void processData(DataQueryProcessExecutionContext context,
			List acquiredData) {

		if (acquiredData == null){
			return;
		}
		ArrayList<BigDecimal> submission = null;

		// each spawned thread to take a <batchsize> list of personId's and construct complete 1095B and submit to IRS delegate
		for (int i = 0; i < acquiredData.size() && !isInterrupted(context); i++) {
			Object[] row = (Object[]) acquiredData.get(i);
			batch.add((BigDecimal)row[0]);
			if (batch.size() % maxBatchSize == 0) {
				submission = new ArrayList<BigDecimal>(batch);
				logger.debug("spawning thread for batch size:" + submission.size());
				processEntityData(context, submission, calendarYear);
				batch.clear();
			}
 		}
		processEntityData(context, batch, calendarYear);
	}

	public void processEntityData(DataQueryProcessExecutionContext context, ArrayList<BigDecimal> submitBatch, String calendarYear) {

		if(isThreaded()) {
			spawnThread(context, submitBatch, calendarYear);
		}

	}

	private void spawnThread(DataQueryProcessExecutionContext context,
			ArrayList<BigDecimal> submitBatch, String calendarYear) {
		try {
			while (getTaskCount(context) >= throttleTaskCountThreshold) {
				Thread.sleep(spawnRetryPeriod);
			}

			// Update the job result once in a while.
			//if(shouldUpdateJobResult(context))
				this.updateJobResult(context);

			initThreadCreatorAndIncrementTaskCount(context);
			ThreadPool threadPool = getThreadPool(context);
			// get a new Thread bean
			GenerateIRSTransmissionProcessSpawnedThreadTask task =
				(GenerateIRSTransmissionProcessSpawnedThreadTask) getApplicationContext().getBean(spawnedTaskId);

			task.setContext(context);
			task.setAcquiredData(submitBatch);
			task.setCalendarYear(calendarYear);
			task.setAuditInfo(getAuditInfo());

			threadPool.invokeLater(task);
		} catch(InterruptedException e) {
				throwIllegalStateException("GenerateIRSTransmissionProcessSpawnedThreadTask was interrupted while it was " +
						"spawning a thread. ", e);
		};
	}

	protected void handleDataProcessCompleted(DataProcessExecutionContext  context) {
		if(isThreaded()) {
			try {
				if(getTaskCount(context) != 0) {
					synchronized(this) {
						boolean stillProcessing = true;
						while(stillProcessing) {
							wait();
							if(getTaskCount(context) == 0)
								stillProcessing = false;
						}
					}
				}
			} catch(InterruptedException e) {
				throwIllegalStateException("GenerateIRSTransmissionProcess was interrupted while it was waiting for " +
						"its spawned threads to complete", e);
			} finally {
				getThreadPool(context).stop();
                cleanThreadPool(context);
			}
		}

		// update the job stats and write the exception data
		this.updateJobResult(context);

       super.handleDataProcessCompleted(context);
	}

	static int getTaskCount(DataProcessExecutionContext context) {
		Integer count = (Integer) context.getContextData().get(CONTEXT_TASK_COUNT);
		return count != null ? count.intValue() : 0;
	}

	private void initThreadCreatorAndIncrementTaskCount(DataQueryProcessExecutionContext context) {
		if(!context.getContextData().containsKey(CONTEXT_THREAD_CREATOR))
			context.getContextData().put(CONTEXT_THREAD_CREATOR, this);

		adjustTaskCount(context, 1);
	}

	static void adjustTaskCount(DataProcessExecutionContext context, int adjustment) {
		synchronized(context) {
			context.getContextData().put(CONTEXT_TASK_COUNT, new Integer(getTaskCount(context) + adjustment));
		}
	}

	private void cleanThreadPool(DataProcessExecutionContext context)
    {
        ThreadPool threadPool = getThreadPool(context);
        threadPool.stop();
        context.getContextData().put(CONTEXT_THREAD_POOL, null);
    }

	private ThreadPool getThreadPool(DataProcessExecutionContext context) {
		ThreadPool threadPool = (ThreadPool) context.getContextData().get(CONTEXT_THREAD_POOL);
		if(threadPool == null) {
			threadPool = new ThreadPool("GenerateIRSTransmissionProcessSpawnedThreadTask", threadPoolSize);
			context.getContextData().put(CONTEXT_THREAD_POOL, threadPool);
		}
		return threadPool;
	}

	protected boolean shouldUpdateJobResult(DataQueryProcessExecutionContext context) {
		return context.getProcessStatistics().isTotalNumberMod(DEFAULT_JOB_RESULT_UPDATE_INTERVAL);
	}


	public SystemParameterService getSystemParameterService() {
		return systemParameterService;
	}

	public void setSystemParameterService(
			SystemParameterService systemParameterService) {
		this.systemParameterService = systemParameterService;
	}

	/**
	 * @return Returns the threadPoolSize.
	 */
	public int getThreadPoolSize() {
		return threadPoolSize;
	}

	public int getFetchSize() {
		return fetchSize;
	}

	public void setFetchSize(int fetchSize) {
		this.fetchSize = fetchSize;
	}

	private boolean isThreaded() {
		return StringUtils.isNotBlank(spawnedTaskId);
	}

	public String getSpawnedTaskId() {
		return spawnedTaskId;
	}

	public void setSpawnedTaskId(String spawnedTaskId) {
		this.spawnedTaskId = spawnedTaskId;
	}

	public int getSpawnRetryPeriod() {
		return spawnRetryPeriod;
	}

	public void setSpawnRetryPeriod(int spawnRetryPeriod) {
		this.spawnRetryPeriod = spawnRetryPeriod;
	}

	/**
	 * @param threadPoolSize The threadPoolSize to set.
	 */
	public void setThreadPoolSize(int threadPoolSize) {
		this.threadPoolSize = threadPoolSize;
	}

	 /*public void afterPropertiesSet() {
			super.afterPropertiesSet();
		}*/

}
