package gov.va.med.esr.common.batchprocess;

import gov.va.med.esr.common.model.CommonEntityKeyFactory;
import gov.va.med.esr.common.model.ee.IRS1095B;
import gov.va.med.esr.common.model.ee.MECPeriod;
import gov.va.med.esr.common.model.person.Person;
import gov.va.med.esr.common.model.person.id.PersonIdEntityKeyImpl;
import gov.va.med.esr.common.persistent.person.MECPeriodDAO;
import gov.va.med.esr.service.IRSTransmissionService;
import gov.va.med.esr.service.PersonMergeService;
import gov.va.med.esr.service.PersonService;
import gov.va.med.fw.batchprocess.DataProcessExecutionContext;
import gov.va.med.fw.batchprocess.DataQueryProcessExecutionContext;
import gov.va.med.fw.persistent.DAOOperations;
import gov.va.med.fw.service.ServiceException;
import gov.va.med.fw.service.support.AbstractSpawnedThreadTask;

import java.math.BigDecimal;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.List;

public class GenerateIRSTransmissionProcessSpawnedThreadTask extends
	AbstractSpawnedThreadTask{

	private IRSTransmissionService irsTransmissionService = null;
	private PersonService personService;
	private ArrayList<BigDecimal> acquiredData;
	private String calendarYear;
	private DataQueryProcessExecutionContext context = null;
	private DAOOperations dao;
	private String BATCH_TYPE = "V";
	private String CORRECTION_IND = "O";
	private MECPeriodDAO mecPeriodDAO = null;
	private PersonMergeService personMergeService;


	/*
	 * (non-Javadoc)
	 *
	 * @see gov.va.med.fw.service.support.SpawnedThreadTask#execute()
	 */
	public void executeTask() throws Throwable {
		//threaded task gets a list of person_id's
		//for each combine any common calyear periods to build single 1095
		BigDecimal personId = null;
		ArrayList<IRS1095B> list = new ArrayList<IRS1095B>();
		IRS1095B vetRecord = null;
		MECPeriod period = null;

		try {
			for (int i = 0; i < acquiredData.size(); i++) {

				try {
					personId = (BigDecimal)acquiredData.get(i);

					if (this.getPersonMergeService().hasDeprecatedRecord(CommonEntityKeyFactory.createPersonIdEntityKey(personId))) {
						logger.error("Skipping Deprecated record found for person: " + personId);
					} else {
						period = new MECPeriod();
						period = combineAllPeriods(personId);
						vetRecord = new IRS1095B();
						vetRecord = build1095(period);

						if (vetRecord == null) {
							logger.error("Failed to build 1095 from person: " + personId);
							context.getProcessStatistics().incrementNumberOfErrorRecords();
						} else {
							list.add(vetRecord);
							context.getProcessStatistics().incrementNumberOfSuccessfulRecords();
						}
					}

				} catch (Exception ex) {
					logger.error("Batch sumission failed to process record: " + personId + " " + ex.getMessage());
					context.getProcessStatistics().incrementNumberOfErrorRecords();
				}
			}
			if (list != null && list.size() > 0) {
				this.getIrsTransmissionService().sendSubmit(list, BATCH_TYPE, CORRECTION_IND, calendarYear);

			}
		} catch (Exception ex) {
			logger.error("Unknown Exception in GenerateIRSSubmission: " + ex.getMessage());
		} finally {
        	GenerateIRSTransmissionProcess.adjustTaskCount(
					this.getContext(), -1);
			Object threadCreator = getThreadCreator(context);
			synchronized (threadCreator) {
				threadCreator.notifyAll();
			}
			//release context for GC of thread pool
			context = null;
		}

	}

	/*
	 * gets all partial periods for the calyear being processed
	 * combines to make one coverage map for the entire year
	 * to avoid sending multiple submissions to IRS with separately covered months,
	 * which is not allowed for original submits
	 */
	private MECPeriod combineAllPeriods(BigDecimal personId) {
		MECPeriod period = new MECPeriod();
		period.setPersonId(personId);
		HashMap coverageMap = new HashMap();
		DateFormat df  = new SimpleDateFormat("MM/dd/yyyy");
		Calendar cal = Calendar.getInstance();

		try {

			List allPeriods = this.getMecPeriodDAO().getByPersonId(personId);

			Calendar start = Calendar.getInstance();
			start.clear();
			start.set(Calendar.YEAR,Integer.valueOf(this.getCalendarYear()));
			start.set(Calendar.MONTH, 0);
		    start.set(Calendar.DATE, 1);

		    if (allPeriods != null) {
		    	for (int i = 0; i < allPeriods.size(); i++) {
		    		MECPeriod partial = (MECPeriod)allPeriods.get(i);

		    		//old period ended before this year, ignore it
		    		if (partial.getMecEndDate() != null && partial.getMecEndDate().before(start.getTime())) {
		    			continue;
		    		}


	    			if (partial.getMecStartDate().before(start.getTime()) && partial.getMecEndDate() == null) {
	    				period.setCoveredAll12Months(true);
	    				period.setCoverageMonths(completeMap());
	    				break;
	    			} else if (partial.getMecStartDate().before(start.getTime())){ //start date last year, end date this year
		    			cal.setTime(partial.getMecStartDate());
	    				int monthStart = 1;
	    				int monthEnd;
	    				if (partial.getMecEndDate() == null) {
	    					monthEnd = 12;
	    				} else {
		    				cal.setTime(partial.getMecEndDate());
		    				monthEnd =  cal.get(Calendar.MONTH) + 1;
	    				}
	    				buildMap(monthStart, monthEnd, coverageMap);
		    		} else { //start date this year, end date this year
		    			cal.setTime(partial.getMecStartDate());
	    				int monthStart = cal.get(Calendar.MONTH) + 1;
	    				int monthEnd;
	    				if (partial.getMecEndDate() == null) {
	    					monthEnd = 12;
	    				} else {
		    				cal.setTime(partial.getMecEndDate());
		    				monthEnd =  cal.get(Calendar.MONTH) + 1;
	    				}
	    				buildMap(monthStart, monthEnd, coverageMap);
		    		}
		    	}
		    }

		    //set all months if all partial periods combine to a full year
		    boolean all = true;
		    for (int i = 1; i <= 12; i++) {
		    	if (!coverageMap.containsKey(Integer.toString(i))) {
		    		all = false;
		    	}
		    }
		    if (all) {
		    	period.setCoveredAll12Months(true);
		    	//coverageMap.clear();
		    }

		    period.setCoverageMonths(coverageMap);
		} catch (Exception ex) {
			logger.error("GenerateIRSSubmitProcess Failed to combine MECPeriods for person: " + personId + " reason:" + ex.getMessage());
		}

		return period;
	}

	private void buildMap(int start, int end, HashMap coverageMap) {
		for (int i = start; i <= end; i++) {
			coverageMap.put(Integer.toString(i), true);
		}
	}

	private HashMap completeMap() {
		HashMap fullYear = new HashMap();
		for (int i = 1; i <= 12; i++) {
			fullYear.put(Integer.toString(i), true);
		}
		return fullYear;
	}

	private IRS1095B build1095(MECPeriod period) {
		IRS1095B rec = null;
		if (period != null) {
			try {
				Person person = personService.getPerson(new PersonIdEntityKeyImpl(period.getPersonId()));
				rec = new IRS1095B(person);
				rec.setMecPeriod(period);
			} catch (ServiceException e) {
				logger.error("Generate IRS Process Failed to Retrieve person by id:" + period.getPersonId() + " , message:" + e.getMessage());
			}
			catch (Exception ex) {
				logger.error("GenerateIRSSubmitProcess Failed to build1095 for person: " + period.getPersonId() + " reason:" + ex.getMessage());
			}
		}
		return rec;
	}


	private Object getThreadCreator(DataProcessExecutionContext context) {
		return context.getContextData().get(GenerateIRSTransmissionProcess.CONTEXT_THREAD_CREATOR);
	}

	/**
	 * @return Returns the acquiredData.
	 */
	public ArrayList getAcquiredData() {
		return acquiredData;
	}

	/**
	 * @param acquiredData
	 *            The acquiredData to set.
	 */
	public void setAcquiredData(ArrayList acquiredData) {
		this.acquiredData = acquiredData;
	}


	/**
	 * @return Returns the personService.
	 */
	public PersonService getPersonService() {
		return this.personService;
	}

	/**
	 * @param personService The personService to set.
	 */
	public void setPersonService(PersonService personService) {
		this.personService = personService;
	}

	/**
	 * @return Returns the dao.
	 */
	public DAOOperations getDao() {
		return dao;
	}

	/**
	 * @param dao
	 * The dao to set.
	 */
	public void setDao(DAOOperations dao) {
		this.dao = dao;
	}

	public MECPeriodDAO getMecPeriodDAO() {
	return mecPeriodDAO;
}

	public void setMecPeriodDAO(MECPeriodDAO mecPeriodDAO) {
		this.mecPeriodDAO = mecPeriodDAO;
	}

	public DataQueryProcessExecutionContext getContext() {
		return context;
	}

	public void setContext(DataQueryProcessExecutionContext context) {
		this.context = context;
	}


	public IRSTransmissionService getIrsTransmissionService() {
		return irsTransmissionService;
	}

	public void setIrsTransmissionService(IRSTransmissionService irsTransmissionService) {
		this.irsTransmissionService = irsTransmissionService;
	}

	public String getCalendarYear() {
		return calendarYear;
	}

	public void setCalendarYear(String calendarYear) {
		this.calendarYear = calendarYear;
	}

	public PersonMergeService getPersonMergeService() {
		return personMergeService;
	}

	public void setPersonMergeService(PersonMergeService personMergeService) {
		this.personMergeService = personMergeService;
	}
}
